@@ -1,34 +1,50 @@
-import asyncio
-from pathlib import Path
-import sys
-import re
-import platform
-import yaml
-import json
-from yaml.parser import ParserError
 import argparse
+import asyncio
+import importlib.metadata
+import json
 import logging
 import os
-from typing import Callable, List
+import platform
+import re
+import sys
+from argparse import ArgumentParser, Namespace
+from pathlib import Path
+from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast
+
 import httpx
-from spectacles import __version__
-from spectacles.runner import Runner
-from spectacles.client import DEFAULT_API_VERSION, LookerClient
+import yaml
+from yaml.parser import ParserError
+
+import spectacles.printer as printer
+from spectacles.client import (
+    DEFAULT_API_VERSION,
+    LOOKML_VALIDATION_TIMEOUT,
+    LookerClient,
+)
 from spectacles.exceptions import (
+    GenericValidationError,
     LookerApiError,
     SpectaclesException,
-    GenericValidationError,
 )
-from spectacles.logger import GLOBAL_LOGGER as logger, set_file_handler
-import spectacles.printer as printer
-import spectacles.tracking as tracking
+from spectacles.logger import GLOBAL_LOGGER as logger
+from spectacles.logger import set_file_handler
+from spectacles.runner import Runner
 from spectacles.utils import log_duration
+from spectacles.validators.data_test import DATA_TEST_CONCURRENCY
+
+__version__ = importlib.metadata.version("spectacles")
 
 
 class ConfigFileAction(argparse.Action):
     """Parses an arbitrary config file and assigns its values as arg defaults."""
 
-    def __call__(self, parser, namespace, values, option_string):
+    def __call__(
+        self,
+        parser: ArgumentParser,
+        namespace: Namespace,
+        values: Union[str, Sequence[Any], None],
+        option_string: Optional[str] = None,
+    ) -> None:
         """Populates argument defaults with values from the config file.
 
         Args:
@@ -38,7 +54,7 @@ class ConfigFileAction(argparse.Action):
             option_string: Argument string, e.g. "--optional".
 
         """
-        config = self.parse_config(path=values)
+        config = self.parse_config(path=values)  # type: ignore
         for dest, value in config.items():
             for action in parser._actions:
                 if dest == action.dest:
@@ -59,7 +75,7 @@ class ConfigFileAction(argparse.Action):
                 )
         parser.set_defaults(**config)
 
-    def parse_config(self, path) -> dict:
+    def parse_config(self, path: str) -> Dict[str, Any]:
         """Base method for parsing an arbitrary config format."""
         raise NotImplementedError()
 
@@ -67,7 +83,7 @@ class ConfigFileAction(argparse.Action):
 class YamlConfigAction(ConfigFileAction):
     """Parses a YAML config file and assigns its values as argument defaults."""
 
-    def parse_config(self, path: str) -> dict:
+    def parse_config(self, path: str) -> Dict[str, Any]:
         """Loads a YAML config file, returning its dictionary format.
 
         Args:
@@ -79,7 +95,7 @@ class YamlConfigAction(ConfigFileAction):
         """
         try:
             with Path(path).open("r") as file:
-                return yaml.safe_load(file)
+                return cast(Dict[str, Any], yaml.safe_load(file))
         except (FileNotFoundError, ParserError) as error:
             raise argparse.ArgumentError(self, str(error))
 
@@ -95,7 +111,13 @@ class EnvVarAction(argparse.Action):
 
     """
 
-    def __init__(self, env_var, required=False, default=None, **kwargs):
+    def __init__(
+        self,
+        env_var: str,
+        required: bool = False,
+        default: Optional[str] = None,
+        **kwargs: Any,
+    ):
         self.env_var = env_var
         if env_var in os.environ:
             default = os.environ[env_var]
@@ -103,7 +125,13 @@ class EnvVarAction(argparse.Action):
             required = False
         super().__init__(default=default, required=required, **kwargs)
 
-    def __call__(self, parser, namespace, values, option_string=None):
+    def __call__(
+        self,
+        parser: ArgumentParser,
+        namespace: Namespace,
+        values: Union[str, Sequence[Any], None],
+        option_string: Optional[str] = None,
+    ) -> None:
         """Sets the argument value to the namespace during parsing.
 
         Args:
@@ -117,7 +145,9 @@ class EnvVarAction(argparse.Action):
 
 
 class EnvVarStoreTrueAction(argparse._StoreTrueAction):
-    def __init__(self, env_var, required=False, default=False, **kwargs):
+    def __init__(
+        self, env_var: str, required: bool = False, default: bool = False, **kwargs: Any
+    ):
         self.env_var = env_var
         if env_var in os.environ:
             value = os.environ[env_var].lower()
@@ -135,11 +165,17 @@ class EnvVarStoreTrueAction(argparse._StoreTrueAction):
             required = False
         super().__init__(default=default, required=required, **kwargs)
 
-    def __call__(self, parser, namespace, values, option_string=None):
+    def __call__(
+        self,
+        parser: ArgumentParser,
+        namespace: Namespace,
+        values: Union[str, Sequence[Any], None],
+        option_string: Optional[str] = None,
+    ) -> None:
         setattr(namespace, self.dest, True)
 
 
-def handle_exceptions(function: Callable) -> Callable:
+def handle_exceptions(function: Callable[..., Any]) -> Callable[..., Any]:
     """Wrapper for handling custom exceptions by logging them.
 
     Args:
@@ -150,7 +186,7 @@ def handle_exceptions(function: Callable) -> Callable:
 
     """
 
-    def wrapper(*args, **kwargs):
+    def wrapper(*args: Any, **kwargs: Any) -> Any:
         try:
             return function(*args, **kwargs)
         except GenericValidationError as error:
@@ -212,24 +248,26 @@ def restore_dash(arg: str) -> str:
     return re.sub(r"^~", "-", arg)
 
 
-def process_pin_imports(input: List[str]) -> dict:
+def process_pin_imports(input: List[str]) -> dict[str, str]:
     return dict(arg.split(":") for arg in input)
 
 
 @handle_exceptions
-def main():
+def main() -> None:
     """Runs main function. This is the entry point."""
-    if sys.version_info < (3, 8):
+    if sys.version_info < (3, 9):
         raise SpectaclesException(
             name="insufficient-python-version",
-            title="Spectacles requires Python 3.8 or higher.",
+            title="Spectacles requires Python 3.9 or higher.",
             detail="The current Python version is %s." % platform.python_version(),
         )
 
     # Convert leading `-` to `~` so they don't break `parse_args`
-    args = [preprocess_dash(arg) for arg in sys.argv[1:]]
+    inputs = [preprocess_dash(arg) for arg in sys.argv[1:]]
     parser = create_parser()
-    args = parser.parse_args(args)
+    args = parser.parse_args(inputs)
+
+    no_looker_ci_warning = getattr(args, "no_looker_ci_warning", False)
 
     branch = getattr(args, "branch", None)
     commit_ref = getattr(args, "commit_ref", None)
@@ -255,13 +293,6 @@ def main():
 
     set_file_handler(args.log_dir)
 
-    if not args.do_not_track:
-        invocation_id = tracking.track_invocation_start(
-            args.base_url,
-            args.command,
-            project=args.project if args.command != "connect" else None,
-        )
-
     if args.command == "connect":
         asyncio.run(
             run_connect(
@@ -294,6 +325,10 @@ def main():
                 chunk_size=args.chunk_size,
                 pin_imports=pin_imports,
                 ignore_hidden=args.ignore_hidden,
+                use_personal_branch=args.use_personal_branch,
+                result_format=(
+                    "json_detail" if args.use_legacy_result_format else "json_bi"
+                ),
             )
         )
     elif args.command == "assert":
@@ -309,6 +344,8 @@ def main():
                 api_version=args.api_version,
                 remote_reset=args.remote_reset,
                 pin_imports=pin_imports,
+                use_personal_branch=args.use_personal_branch,
+                concurrency=args.concurrency,
             )
         )
     elif args.command == "content":
@@ -328,6 +365,7 @@ def main():
                 exclude_personal=args.exclude_personal,
                 folders=[restore_dash(arg) for arg in args.folders],
                 pin_imports=pin_imports,
+                use_personal_branch=args.use_personal_branch,
             )
         )
     elif args.command == "lookml":
@@ -343,26 +381,24 @@ def main():
                 remote_reset=args.remote_reset,
                 severity=args.severity,
                 pin_imports=pin_imports,
+                use_personal_branch=args.use_personal_branch,
+                timeout=args.timeout,
             )
         )
 
-    if not args.do_not_track:
-        tracking.track_invocation_end(
-            args.base_url,
-            args.command,
-            invocation_id,  # pyright: ignore[reportUnboundVariable]
-            args.project if args.command != "connect" else None,
-        )
+    # print out announcement about Looker CI Public Preview
+    if not no_looker_ci_warning:
+        printer.print_looker_ci_warning()
 
 
-def create_parser() -> argparse.ArgumentParser:
+def create_parser() -> ArgumentParser:
     """Creates the top-level argument parser.
 
     Returns:
-        argparse.ArgumentParser: Top-level argument parser.
+        ArgumentParser: Top-level argument parser.
 
     """
-    parser = argparse.ArgumentParser(prog="spectacles")
+    parser = ArgumentParser(prog="spectacles")
     parser.add_argument("--version", action="version", version=__version__)
     subparser_action = parser.add_subparsers(
         title="Available sub-commands", dest="command"
@@ -376,14 +412,14 @@ def create_parser() -> argparse.ArgumentParser:
     return parser
 
 
-def _build_base_subparser() -> argparse.ArgumentParser:
+def _build_base_subparser() -> ArgumentParser:
     """Returns the base subparser with arguments required for every subparser.
 
     Returns:
-        argparse.ArgumentParser: Base subparser with url and auth arguments.
+        ArgumentParser: Base subparser with url and auth arguments.
 
     """
-    base_subparser = argparse.ArgumentParser(add_help=False)
+    base_subparser = ArgumentParser(add_help=False)
     base_subparser.add_argument(
         "--config-file",
         action=YamlConfigAction,
@@ -443,19 +479,25 @@ def _build_base_subparser() -> argparse.ArgumentParser:
         default="logs",
         help="The directory that Spectacles will write logs to.",
     )
+    base_subparser.add_argument(
+        "--no-looker-ci-warning",
+        action=EnvVarStoreTrueAction,
+        env_var="SPECTACLES_NO_LOOKER_CI_WARNING",
+        help="Flag to suppress the announcement about Looker CI's Public Preview.",
+    )
     base_subparser.add_argument(
         "--do-not-track",
         action=EnvVarStoreTrueAction,
         env_var="SPECTACLES_DO_NOT_TRACK",
-        help="Disables anonymised event tracking.",
+        help="[DEPRECATED] Tracking is disabled by default, this argument is unused.",
     )
 
     return base_subparser
 
 
 def _build_connect_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
 ) -> None:
     """Returns the subparser for the subcommand `connect`.
 
@@ -475,13 +517,13 @@ def _build_connect_subparser(
 
 
 def _build_validator_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
-) -> argparse.ArgumentParser:
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
+) -> ArgumentParser:
     """Returns the base subparser with arguments required for every validator.
 
     Returns:
-        argparse.ArgumentParser: validator subparser with project, branch, remote reset and import projects arguments.
+        ArgumentParser: validator subparser with project, branch, remote reset and import projects arguments.
 
     """
     base_subparser.add_argument(
@@ -497,6 +539,19 @@ def _build_validator_subparser(
         env_var="LOOKER_GIT_BRANCH",
         help="The branch of your project that Spectacles will use to run queries.",
     )
+    base_subparser.add_argument(
+        "--use-personal-branch",
+        action=EnvVarStoreTrueAction,
+        env_var="SPECTACLES_USE_PERSONAL_BRANCH",
+        help="Use the user's personal branch instead of creating a temporary branch for the tests.",
+    )
+    base_subparser.add_argument(
+        "--pin-imports",
+        nargs="+",
+        default=[],
+        help="Pin locally imported Looker projects to a specific ref (Git branch or commit) during validation. \
+            Provide these arguments in project_name:ref format.",
+    )
     group = base_subparser.add_mutually_exclusive_group()
     group.add_argument(
         "--remote-reset",
@@ -514,20 +569,13 @@ def _build_validator_subparser(
             In order to test a specific commit, Spectacles will create a new branch \
             for the tests and then delete the branch when it is finished.",
     )
-    group.add_argument(
-        "--pin-imports",
-        nargs="+",
-        default=[],
-        help="Pin locally imported Looker projects to a specific ref (Git branch or commit) during validation. \
-            Provide these arguments in project_name:ref format.",
-    )
     return base_subparser
 
 
 def _build_select_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
-) -> argparse.ArgumentParser:
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
+) -> ArgumentParser:
     base_subparser.add_argument(
         "--explores",
         nargs="+",
@@ -541,8 +589,8 @@ def _build_select_subparser(
 
 
 def _build_lookml_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
 ) -> None:
     """Returns the subparser for the subcommand `lookml`.
 
@@ -570,12 +618,18 @@ def _build_lookml_subparser(
             "validator to fail. The default is 'warning'."
         ),
     )
+    subparser.add_argument(
+        "--timeout",
+        type=int,
+        default=LOOKML_VALIDATION_TIMEOUT,
+        help="Specify the timeout for the LookML validation in seconds.",
+    )
     _build_validator_subparser(subparser_action, subparser)
 
 
 def _build_sql_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
 ) -> None:
     """Returns the subparser for the subcommand `sql`.
 
@@ -630,7 +684,8 @@ def _build_sql_subparser(
         action="store_true",
         help=(
             "After validation, display queries that took longer than the runtime "
-            "threshold (5 seconds by default) to complete."
+            "threshold (5 seconds by default) to complete. Must use the legacy "
+            "result format (--use-legacy-result-format) if enabling profiler."
         ),
     )
     subparser.add_argument(
@@ -653,13 +708,18 @@ def _build_sql_subparser(
         action="store_true",
         help=("Exclude hidden fields from validation."),
     )
+    subparser.add_argument(
+        "--use-legacy-result-format",
+        action="store_true",
+        help="Use the deprecated json_detail result format when creating queries with the Looker API (instead of the default, json_bi).",
+    )
     _build_validator_subparser(subparser_action, subparser)
     _build_select_subparser(subparser_action, subparser)
 
 
 def _build_assert_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
 ) -> None:
     """Returns the subparser for the subcommand `assert`.
 
@@ -677,10 +737,20 @@ def _build_assert_subparser(
     _build_validator_subparser(subparser_action, subparser)
     _build_select_subparser(subparser_action, subparser)
 
+    subparser.add_argument(
+        "--concurrency",
+        type=int,
+        default=DATA_TEST_CONCURRENCY,
+        help=(
+            "Specify the number of concurrent queries you want to have running "
+            f"against your data warehouse. The default is {DATA_TEST_CONCURRENCY}."
+        ),
+    )
+
 
 def _build_content_subparser(
-    subparser_action: argparse._SubParsersAction,
-    base_subparser: argparse.ArgumentParser,
+    subparser_action: argparse._SubParsersAction,  # type: ignore[type-arg]
+    base_subparser: ArgumentParser,
 ) -> None:
     subparser = subparser_action.add_parser(
         "content", parents=[base_subparser], help="Run Looker content validation."
@@ -737,16 +807,18 @@ async def run_connect(
 
 @log_duration
 async def run_lookml(
-    project,
-    ref,
-    base_url,
-    client_id,
-    client_secret,
-    port,
-    api_version,
-    remote_reset,
-    severity,
-    pin_imports,
+    project: str,
+    ref: str,
+    base_url: str,
+    client_id: str,
+    client_secret: str,
+    port: int,
+    api_version: float,
+    remote_reset: bool,
+    severity: str,
+    pin_imports: Dict[str, str],
+    use_personal_branch: bool,
+    timeout: int,
 ) -> None:
     # Don't trust env to ignore .netrc credentials
     async_client = httpx.AsyncClient(trust_env=False)
@@ -754,9 +826,9 @@ async def run_lookml(
         client = LookerClient(
             async_client, base_url, client_id, client_secret, port, api_version
         )
-        runner = Runner(client, project, remote_reset, pin_imports)
+        runner = Runner(client, project, remote_reset, pin_imports, use_personal_branch)
 
-        results = await runner.validate_lookml(ref, severity)
+        results = await runner.validate_lookml(ref, severity, timeout)
     finally:
         await async_client.aclose()
 
@@ -791,20 +863,21 @@ async def run_lookml(
 
 @log_duration
 async def run_content(
-    project,
-    ref,
-    filters,
-    base_url,
-    client_id,
-    client_secret,
-    port,
-    api_version,
-    remote_reset,
-    incremental,
-    target,
-    exclude_personal,
-    folders,
-    pin_imports,
+    project: str,
+    ref: str,
+    filters: List[str],
+    base_url: str,
+    client_id: str,
+    client_secret: str,
+    port: int,
+    api_version: float,
+    remote_reset: bool,
+    incremental: bool,
+    target: str,
+    exclude_personal: bool,
+    folders: List[str],
+    pin_imports: Dict[str, str],
+    use_personal_branch: bool,
 ) -> None:
     # Don't trust env to ignore .netrc credentials
     async_client = httpx.AsyncClient(trust_env=False)
@@ -812,7 +885,7 @@ async def run_content(
         client = LookerClient(
             async_client, base_url, client_id, client_secret, port, api_version
         )
-        runner = Runner(client, project, remote_reset, pin_imports)
+        runner = Runner(client, project, remote_reset, pin_imports, use_personal_branch)
 
         results = await runner.validate_content(
             ref,
@@ -854,16 +927,18 @@ async def run_content(
 
 @log_duration
 async def run_assert(
-    project,
-    ref,
-    filters,
-    base_url,
-    client_id,
-    client_secret,
-    port,
-    api_version,
-    remote_reset,
-    pin_imports,
+    project: str,
+    ref: str,
+    filters: List[str],
+    base_url: str,
+    client_id: str,
+    client_secret: str,
+    port: int,
+    api_version: float,
+    remote_reset: bool,
+    pin_imports: Dict[str, str],
+    use_personal_branch: bool,
+    concurrency: int,
 ) -> None:
     # Don't trust env to ignore .netrc credentials
     async_client = httpx.AsyncClient(trust_env=False)
@@ -871,9 +946,9 @@ async def run_assert(
         client = LookerClient(
             async_client, base_url, client_id, client_secret, port, api_version
         )
-        runner = Runner(client, project, remote_reset, pin_imports)
+        runner = Runner(client, project, remote_reset, pin_imports, use_personal_branch)
 
-        results = await runner.validate_data_tests(ref, filters)
+        results = await runner.validate_data_tests(ref, filters, concurrency)
     finally:
         await async_client.aclose()
 
@@ -906,25 +981,27 @@ async def run_assert(
 
 @log_duration
 async def run_sql(
-    log_dir,
-    project,
-    ref,
-    filters,
-    base_url,
-    client_id,
-    client_secret,
-    port,
-    api_version,
-    fail_fast,
-    incremental,
-    target,
-    remote_reset,
-    concurrency,
-    profile,
-    runtime_threshold,
-    chunk_size,
-    pin_imports,
-    ignore_hidden,
+    log_dir: str,
+    project: str,
+    ref: str,
+    filters: List[str],
+    base_url: str,
+    client_id: str,
+    client_secret: str,
+    port: int,
+    api_version: float,
+    fail_fast: bool,
+    incremental: bool,
+    target: str,
+    remote_reset: bool,
+    concurrency: int,
+    profile: bool,
+    runtime_threshold: int,
+    chunk_size: int,
+    pin_imports: Dict[str, str],
+    use_personal_branch: bool,
+    ignore_hidden: bool,
+    result_format: str,
 ) -> None:
     """Runs and validates the SQL for each selected LookML dimension."""
     # Don't trust env to ignore .netrc credentials
@@ -933,7 +1010,7 @@ async def run_sql(
         client = LookerClient(
             async_client, base_url, client_id, client_secret, port, api_version
         )
-        runner = Runner(client, project, remote_reset, pin_imports)
+        runner = Runner(client, project, remote_reset, pin_imports, use_personal_branch)
 
         results = await runner.validate_sql(
             ref,
@@ -946,13 +1023,16 @@ async def run_sql(
             runtime_threshold,
             chunk_size,
             ignore_hidden,
+            result_format,
         )
     finally:
         await async_client.aclose()
 
     for test in sorted(results["tested"], key=lambda x: (x["model"], x["explore"])):
         message = f"{test['model']}.{test['explore']}"
-        printer.print_validation_result(status=test["status"], source=message)
+        printer.print_validation_result(
+            status=test["status"], skip_reason=test.get("skip_reason"), source=message
+        )
 
     errors = sorted(
         results["errors"],
