@@ -1,24 +1,59 @@
-from typing import List, Dict, Optional, Tuple
+import asyncio
+import json
 import time
 from dataclasses import dataclass
-import backoff  # type: ignore
+from http import HTTPStatus
+from typing import Any, Callable, Dict, List, Optional, Tuple
+
+import backoff
 import httpx
-from httpx import HTTPStatusError, ConnectError, TimeoutException
+from aiocache import Cache, cached, serializers
+from httpx import (
+    HTTPStatusError,
+    NetworkError,
+    ReadTimeout,
+    RemoteProtocolError,
+    TimeoutException,
+)
+
 import spectacles.utils as utils
-from spectacles.types import JsonDict
+from spectacles.exceptions import LookerApiError, SpectaclesException
 from spectacles.logger import GLOBAL_LOGGER as logger
-from spectacles.exceptions import SpectaclesException, LookerApiError
+from spectacles.models import JsonDict
 
 DEFAULT_API_VERSION = 4.0
 TIMEOUT_SEC = 300
+LOOKML_VALIDATION_TIMEOUT = 7200
 MAX_ASYNC_CONNECTIONS = 200
-DEFAULT_MAX_TRIES = 3
-BACKOFF_EXCEPTIONS = (
+
+DEFAULT_RETRIES = 3
+DEFAULT_NETWORK_RETRIES = 10
+NETWORK_EXCEPTIONS = (
+    NetworkError,
     TimeoutException,
+    RemoteProtocolError,
+)
+STATUS_EXCEPTIONS = (
     HTTPStatusError,
-    ConnectError,
     LookerApiError,
 )
+BACKOFF_EXCEPTIONS = NETWORK_EXCEPTIONS + STATUS_EXCEPTIONS
+
+
+def giveup_unless_bad_gateway(exception: Exception) -> bool:
+    """Give up retries if a status error encountered with any code besides 502/504."""
+    if isinstance(exception, LookerApiError):
+        return exception.status not in (
+            HTTPStatus.BAD_GATEWAY,
+            HTTPStatus.GATEWAY_TIMEOUT,
+        )
+    elif isinstance(exception, HTTPStatusError):
+        return exception.response.status_code not in (
+            HTTPStatus.BAD_GATEWAY,
+            HTTPStatus.GATEWAY_TIMEOUT,
+        )
+    else:
+        return False
 
 
 @dataclass(frozen=True)  # Token is immutable
@@ -37,6 +72,27 @@ class AccessToken:
         return False if time.time() < self.expires_at else True
 
 
+def backoff_with_exceptions(func: Callable[..., Any]) -> Callable[..., Any]:
+    @backoff.on_exception(
+        backoff.expo,
+        STATUS_EXCEPTIONS,
+        giveup=giveup_unless_bad_gateway,
+        max_tries=DEFAULT_RETRIES,
+    )
+    @backoff.on_exception(
+        backoff.expo,
+        NETWORK_EXCEPTIONS,
+        max_tries=DEFAULT_NETWORK_RETRIES,
+    )
+    async def wrapper(*args: Any, **kwargs: Any) -> Any:
+        if asyncio.iscoroutinefunction(func):
+            return await func(*args, **kwargs)
+        else:
+            return func(*args, **kwargs)
+
+    return wrapper
+
+
 class LookerClient:
     """Wraps some endpoints of the Looker API, issues requests and handles responses.
 
@@ -139,7 +195,9 @@ class LookerClient:
             f"using Looker API {self.api_version}"
         )
 
-    async def request(self, method: str, url: str, *args, **kwargs) -> httpx.Response:
+    async def request(
+        self, method: str, url: str, *args: Any, **kwargs: Any
+    ) -> httpx.Response:
         if self.access_token and self.access_token.expired:
             logger.debug("Looker API access token has expired, requesting a new one")
             self.authenticate()
@@ -147,22 +205,27 @@ class LookerClient:
                 await self.update_workspace("dev")
         return await self.async_client.request(method, url, *args, **kwargs)
 
-    async def get(self, url, *args, **kwargs) -> httpx.Response:
+    async def get(self, url: str, *args: Any, **kwargs: Any) -> httpx.Response:
         return await self.request("GET", url, *args, **kwargs)
 
-    async def post(self, url, *args, **kwargs) -> httpx.Response:
+    async def post(self, url: str, *args: Any, **kwargs: Any) -> httpx.Response:
         return await self.request("POST", url, *args, **kwargs)
 
-    async def patch(self, url, *args, **kwargs) -> httpx.Response:
+    async def patch(self, url: str, *args: Any, **kwargs: Any) -> httpx.Response:
         return await self.request("PATCH", url, *args, **kwargs)
 
-    async def put(self, url, *args, **kwargs) -> httpx.Response:
+    async def put(self, url: str, *args: Any, **kwargs: Any) -> httpx.Response:
         return await self.request("PUT", url, *args, **kwargs)
 
-    async def delete(self, url, *args, **kwargs) -> httpx.Response:
+    async def delete(self, url: str, *args: Any, **kwargs: Any) -> httpx.Response:
         return await self.request("DELETE", url, *args, **kwargs)
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff.on_exception(
+        backoff.expo,
+        BACKOFF_EXCEPTIONS,
+        giveup=giveup_unless_bad_gateway,
+        max_tries=DEFAULT_NETWORK_RETRIES,
+    )
     def get_looker_release_version(self) -> str:
         """Gets the version number of connected Looker instance.
 
@@ -191,9 +254,9 @@ class LookerClient:
                 response=response,
             ) from error
 
-        return response.json()["looker_release_version"]
+        return response.json()["looker_release_version"]  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def get_workspace(self) -> str:
         """Gets the session workspace.
 
@@ -219,9 +282,9 @@ class LookerClient:
                 ),
                 response=response,
             ) from error
-        return response.json()["workspace_id"]
+        return response.json()["workspace_id"]  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def update_workspace(self, workspace: str) -> None:
         """Updates the session workspace.
 
@@ -249,8 +312,8 @@ class LookerClient:
             ) from error
         self.workspace = workspace
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def get_all_branches(self, project: str) -> List[str]:
+    @backoff_with_exceptions
+    async def get_all_branches(self, project: str) -> List[JsonDict]:
         """Returns a list of git branches in the project repository.
 
         Args:
@@ -275,9 +338,9 @@ class LookerClient:
                 response=response,
             ) from error
 
-        return [branch["name"] for branch in response.json()]
+        return response.json()  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def checkout_branch(self, project: str, branch: str) -> None:
         """Checks out a new git branch. Only works in dev workspace.
 
@@ -304,7 +367,7 @@ class LookerClient:
                 response=response,
             ) from error
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def reset_to_remote(self, project: str) -> None:
         """Reset a project development branch to the revision of the project that is on the remote.
 
@@ -331,6 +394,7 @@ class LookerClient:
                 response=response,
             ) from error
 
+    @backoff_with_exceptions
     async def get_manifest(self, project: str) -> JsonDict:
         """Gets all the dependent LookML projects defined in the manifest file.
 
@@ -361,9 +425,9 @@ class LookerClient:
 
         manifest = response.json()
 
-        return manifest
+        return manifest  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def get_active_branch(self, project: str) -> JsonDict:
         """Gets the active branch for the user in the given project.
 
@@ -386,8 +450,8 @@ class LookerClient:
                 status=response.status_code,
                 detail=(
                     f"Unable to get active branch for project '{project}'. "
-                    "Please check that the project exists and that your user "
-                    "has the correct permissions and try again."
+                    "Please check that the project exists, is configured, "
+                    "and that your user has the correct permissions and try again."
                 ),
                 response=response,
             ) from error
@@ -395,15 +459,15 @@ class LookerClient:
         branch_name = response.json()["name"]
         logger.debug(f"The active branch is '{branch_name}'")
 
-        return response.json()
+        return response.json()  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def get_active_branch_name(self, project: str) -> str:
         """Helper method to return only the branch name."""
         full_response = await self.get_active_branch(project)
-        return full_response["name"]
+        return full_response["name"]  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def create_branch(
         self, project: str, branch: str, ref: Optional[str] = None
     ) -> None:
@@ -447,7 +511,7 @@ class LookerClient:
                 response=response,
             ) from error
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def hard_reset_branch(self, project: str, branch: str, ref: str) -> None:
         """Hard resets a branch to the ref prodvided.
 
@@ -482,8 +546,8 @@ class LookerClient:
                 response=response,
             ) from error
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def delete_branch(self, project: str, branch: str):
+    @backoff_with_exceptions
+    async def delete_branch(self, project: str, branch: str) -> None:
         """Deletes a branch in the given project.
 
         Args:
@@ -511,7 +575,7 @@ class LookerClient:
                 response=response,
             ) from error
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def all_lookml_tests(self, project: str) -> List[JsonDict]:
         """Gets all LookML/data tests for a given project.
 
@@ -542,9 +606,9 @@ class LookerClient:
                 response=response,
             ) from error
 
-        return response.json()
+        return response.json()  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def run_lookml_test(
         self, project: str, model: Optional[str] = None, test: Optional[str] = None
     ) -> List[JsonDict]:
@@ -599,10 +663,12 @@ class LookerClient:
                 response=response,
             ) from error
 
-        return response.json()
+        return response.json()  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def get_lookml_models(self, fields: Optional[List] = None) -> List[JsonDict]:
+    @backoff_with_exceptions
+    async def get_lookml_models(
+        self, fields: Optional[List[str]] = None
+    ) -> List[JsonDict]:
         """Gets all models and explores from the LookmlModel endpoint.
 
         Returns:
@@ -613,7 +679,7 @@ class LookerClient:
         if fields is None:
             fields = []
 
-        params = {}
+        params: Dict[str, Any] = {}
         if fields:
             params["fields"] = fields
 
@@ -630,21 +696,13 @@ class LookerClient:
                 response=response,
             ) from error
 
-        return response.json()
+        return response.json()  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def get_lookml_dimensions(self, model: str, explore: str) -> List[str]:
-        """Gets all dimensions for an explore from the LookmlModel endpoint.
-
-        Args:
-            model: Name of LookML model to query.
-            explore: Name of LookML explore to query.
-
-        Returns:
-            List[str]: Names of all the dimensions in the specified explore. Dimension
-                names are returned in the format 'explore_name.dimension_name'.
-
-        """
+    @backoff_with_exceptions
+    async def get_lookml_dimensions(
+        self, model: str, explore: str
+    ) -> List[Dict[str, Any]]:
+        """Gets all dimensions for an explore from the LookmlModel endpoint."""
         logger.debug(f"Getting all dimensions from explore {model}/{explore}")
         params = {"fields": ["fields"]}
         url = utils.compose_url(
@@ -667,16 +725,17 @@ class LookerClient:
                 response=response,
             ) from error
 
-        return response.json()["fields"]["dimensions"]
+        return response.json()["fields"]["dimensions"]  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=5)
+    @cached(cache=Cache.MEMORY, serializer=serializers.PickleSerializer())  # type: ignore
+    @backoff_with_exceptions
     async def create_query(
         self,
         model: str,
         explore: str,
         dimensions: List[str],
-        fields: Optional[List] = None,
-    ) -> Dict:
+        fields: Optional[List[str]] = None,
+    ) -> JsonDict:
         """Creates a Looker async query for one or more specified dimensions.
 
         The query created is a SELECT query, selecting all dimensions specified for a
@@ -701,7 +760,7 @@ class LookerClient:
             "filter_expression": "1=2",
         }
 
-        params: Dict[str, list] = {}
+        params: Dict[str, List[str]] = {}
         if fields is None:
             params["fields"] = []
         else:
@@ -733,10 +792,12 @@ class LookerClient:
             "*" if len(dimensions) != 1 else dimensions[0],
             query_id,
         )
-        return result
+        return result  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=5)
-    async def create_query_task(self, query_id: str) -> str:
+    @backoff_with_exceptions
+    async def create_query_task(
+        self, query_id: str, result_format: str = "json_bi"
+    ) -> str:
         """Runs a previously created query asynchronously and returns the query task ID.
 
         If a ClientError or TimeoutError is received, attempts to retry.
@@ -752,7 +813,7 @@ class LookerClient:
         """
         # Using old-style string formatting so that strings are formatted lazily
         logger.debug("Starting query %s", query_id)
-        body = {"query_id": query_id, "result_format": "json_detail"}
+        body = {"query_id": query_id, "result_format": result_format}
         params = {"fields": ["id"]}
         url = utils.compose_url(self.api_url, path=["query_tasks"], params=params)
 
@@ -777,9 +838,9 @@ class LookerClient:
         result = response.json()
         query_task_id = result["id"]
         logger.debug("Query %s is running under query task %s", query_id, query_task_id)
-        return query_task_id
+        return query_task_id  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def get_query_task_multi_results(
         self, query_task_ids: Tuple[str, ...]
     ) -> JsonDict:
@@ -821,9 +882,9 @@ class LookerClient:
             ) from error
 
         result = response.json()
-        return result
+        return result  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @backoff_with_exceptions
     async def cancel_query_task(self, query_task_id: str) -> None:
         """Cancels a query task.
 
@@ -857,13 +918,28 @@ class LookerClient:
             ) from error
 
         result = response.json()
-        return result
+        return result  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def lookml_validation(self, project) -> JsonDict:
-        logger.debug(f"Validating LookML for project '{project}'")
+    @backoff.on_exception(
+        backoff.expo,
+        # Omit retries on timeouts because timeout is already very long
+        STATUS_EXCEPTIONS
+        + (
+            NetworkError,
+            RemoteProtocolError,
+            ReadTimeout,
+        ),
+        giveup=giveup_unless_bad_gateway,
+        max_tries=DEFAULT_RETRIES,
+    )
+    async def lookml_validation(
+        self, project: str, timeout: int = LOOKML_VALIDATION_TIMEOUT
+    ) -> JsonDict:
+        logger.debug(
+            f"Validating LookML for project '{project}' with timeout {timeout} seconds."
+        )
         url = utils.compose_url(self.api_url, path=["projects", project, "validate"])
-        response = await self.post(url=url, timeout=1800)
+        response = await self.post(url=url, timeout=timeout)
 
         try:
             response.raise_for_status()
@@ -877,10 +953,20 @@ class LookerClient:
             ) from error
 
         result = response.json()
-        return result
+        return result  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def cached_lookml_validation(self, project) -> Optional[JsonDict]:
+    @backoff.on_exception(
+        backoff.expo,
+        # Omit retries on timeouts because timeout is already very long
+        STATUS_EXCEPTIONS
+        + (
+            NetworkError,
+            RemoteProtocolError,
+        ),
+        giveup=giveup_unless_bad_gateway,
+        max_tries=DEFAULT_RETRIES,
+    )
+    async def cached_lookml_validation(self, project: str) -> Optional[JsonDict]:
         logger.debug(f"Getting cached LookML validation results for '{project}'")
         url = utils.compose_url(self.api_url, path=["projects", project, "validate"])
         response = await self.get(url=url, timeout=1800)
@@ -904,9 +990,10 @@ class LookerClient:
             return None
 
         result = response.json()
-        return result
+        return result  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
+    @cached(cache=Cache.MEMORY, serializer=serializers.PickleSerializer())  # type: ignore
+    @backoff_with_exceptions
     async def all_folders(self) -> List[JsonDict]:
         logger.debug("Getting information about all folders")
         url = utils.compose_url(self.api_url, path=["folders"])
@@ -924,10 +1011,12 @@ class LookerClient:
             ) from error
 
         result = response.json()
-        return result
+        return result  # type: ignore[no-any-return]
 
-    @backoff.on_exception(backoff.expo, BACKOFF_EXCEPTIONS, max_tries=DEFAULT_MAX_TRIES)
-    async def run_query(self, query_id: str) -> str:
+    @backoff_with_exceptions
+    async def run_query(
+        self, query_id: str, explore: str, model: str, dimension: Optional[str] = None
+    ) -> str:
         """Returns the compiled SQL for a given query ID.
 
         The corresponding Looker API endpoint allows us to run queries with a variety
@@ -949,15 +1038,33 @@ class LookerClient:
                 return (
                     "-- SQL could not be generated because of errors with this query."
                 )
+            elif e.response.status_code == 400 and (
+                "Must query at least one dimension or measure" in response.text
+                or "View Not Found" in response.text
+            ):
+                return (
+                    "-- SQL could not be generated because of errors with this query."
+                )
             else:
+                detail = (
+                    f"Failed to retrieve compiled SQL for "
+                    f"{'dimension' if dimension else 'explore'} "
+                    f"'{model}/{explore}{'/' + dimension if dimension else ''}' "
+                    f"with query '{query_id}'. "
+                    "Please try again."
+                )
+                try:
+                    response_json = response.json()
+                    if "message" in response_json:
+                        message = response_json["message"]
+                        detail += f' Received the following from Looker: "{message}"'
+                except json.JSONDecodeError:
+                    pass
                 raise LookerApiError(
                     name="unable-to-retrieve-compiled-sql",
                     title="Couldn't retrieve compiled SQL.",
                     status=response.status_code,
-                    detail=(
-                        f"Failed to retrieve compiled SQL for query '{query_id}'. "
-                        "Please try again."
-                    ),
+                    detail=detail,
                     response=response,
                 ) from e
 
@@ -965,3 +1072,54 @@ class LookerClient:
         logger.debug("Retrieved compiled SQL for query %s", query_id)
 
         return result
+
+    @backoff_with_exceptions
+    async def run_inline_query(
+        self,
+        query_body: Dict[str, Any],
+        result_format: str,
+        model: str,
+        explore: str,
+        dimension: Optional[str] = None,
+    ) -> str:
+        """Runs a query inline and returns the result in the specified format."""
+        logger.debug(f"Running inline query for {model}/{explore}")
+        url = utils.compose_url(self.api_url, path=["queries", "run", result_format])
+        response = await self.post(url=url, json=query_body, timeout=TIMEOUT_SEC)
+        try:
+            response.raise_for_status()
+        except httpx.HTTPStatusError as e:
+            if e.response.status_code == 404:
+                return (
+                    "-- SQL could not be generated because of errors with this query."
+                )
+            elif e.response.status_code == 400 and (
+                "Must query at least one dimension or measure" in response.text
+                or "View Not Found" in response.text
+            ):
+                return (
+                    "-- SQL could not be generated because of errors with this query."
+                )
+            else:
+                detail = (
+                    f"Failed to retrieve compiled SQL for "
+                    f"{'dimension' if dimension else 'explore'} "
+                    f"'{model}/{explore}{'/' + dimension if dimension else ''}'. "
+                    "Please try again."
+                )
+                try:
+                    response_json = response.json()
+                    if "message" in response_json:
+                        message = response_json["message"]
+                        detail += f' Received the following from Looker: "{message}"'
+                except json.JSONDecodeError:
+                    pass
+                raise LookerApiError(
+                    name="unable-to-get-sql",
+                    title="Couldn't get compiled SQL.",
+                    status=e.response.status_code,
+                    detail=detail,
+                    response=e.response,
+                ) from e
+
+        return response.text
