@@ -1,15 +1,29 @@
+import asyncio
 import re
-from typing import Dict, List, Sequence, Optional, Any, Iterable
+from dataclasses import dataclass
+from typing import Any, Dict, Generator, Iterable, List, Optional, Sequence
+
 from spectacles.client import LookerClient
-from spectacles.exceptions import ValidationError, LookMlNotFound
-from spectacles.types import JsonDict
-from spectacles.select import is_selected
+from spectacles.exceptions import LookMlNotFound, ValidationError
+from spectacles.logger import GLOBAL_LOGGER as logger
+from spectacles.models import JsonDict, SkipReason
+from spectacles.project_select import is_selected
 
 
 class LookMlObject:
-    def __repr__(self):
+    name: str
+
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}(name={self.name})"
 
+    @property
+    def queried(self) -> bool:
+        raise NotImplementedError
+
+    @queried.setter
+    def queried(self, value: bool) -> None:
+        raise NotImplementedError
+
 
 class Dimension(LookMlObject):
     def __init__(
@@ -31,7 +45,7 @@ class Dimension(LookMlObject):
         self.sql = sql
         self.url = url
         self.is_hidden = is_hidden
-        self.queried: bool = False
+        self._queried: bool = False
         self.errors: List[ValidationError] = []
 
         if (
@@ -42,14 +56,22 @@ class Dimension(LookMlObject):
         else:
             self.ignore = False
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             f"{self.__class__.__name__}(name={self.name}, "
             f"type={self.type}, "
             f"errored={self.errored})"
         )
 
-    def __eq__(self, other):
+    @property
+    def queried(self) -> bool:
+        return self._queried
+
+    @queried.setter
+    def queried(self, value: bool) -> None:
+        self._queried = value
+
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, Dimension):
             return NotImplemented
 
@@ -61,12 +83,22 @@ class Dimension(LookMlObject):
             and self.url == other.url
         )
 
+    def __lt__(self, other: Any) -> bool:
+        if not isinstance(other, Dimension):
+            return NotImplemented
+
+        return (self.model_name, self.explore_name, self.name) < (
+            other.model_name,
+            other.explore_name,
+            other.name,
+        )
+
     @property
-    def errored(self):
+    def errored(self) -> Optional[bool]:
         return bool(self.errors) if self.queried else None
 
     @errored.setter
-    def errored(self, value):
+    def errored(self, value: bool) -> None:
         raise AttributeError(
             "Cannot assign to 'errored' property of a Dimension instance. "
             "For a dimension to be considered errored, it must have a ValidationError "
@@ -74,7 +106,9 @@ class Dimension(LookMlObject):
         )
 
     @classmethod
-    def from_json(cls, json_dict, model_name, explore_name):
+    def from_json(
+        cls, json_dict: Dict[str, Any], model_name: str, explore_name: str
+    ) -> "Dimension":
         name = json_dict["name"]
         type = json_dict["type"]
         tags = json_dict["tags"]
@@ -85,16 +119,18 @@ class Dimension(LookMlObject):
 
 
 class Explore(LookMlObject):
-    def __init__(self, name: str, model_name: str, dimensions: List[Dimension] = None):
+    def __init__(
+        self, name: str, model_name: str, dimensions: Optional[List[Dimension]] = None
+    ):
         self.name = name
         self.model_name = model_name
         self.dimensions = [] if dimensions is None else dimensions
         self.errors: List[ValidationError] = []
         self.successes: List[JsonDict] = []
-        self.skipped = False
+        self.skipped: Optional[SkipReason] = None
         self._queried: bool = False
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, Explore):
             return NotImplemented
 
@@ -105,14 +141,14 @@ class Explore(LookMlObject):
         )
 
     @property
-    def queried(self):
+    def queried(self) -> bool:
         if self.dimensions:
             return any(dimension.queried for dimension in self.dimensions)
         else:
             return self._queried
 
     @queried.setter
-    def queried(self, value: bool):
+    def queried(self, value: bool) -> None:
         if not isinstance(value, bool):
             raise TypeError("Value for queried must be boolean.")
         if self.dimensions:
@@ -122,7 +158,7 @@ class Explore(LookMlObject):
             self._queried = value
 
     @property
-    def errored(self):
+    def errored(self) -> Optional[bool]:
         if self.queried:
             return bool(self.errors) or any(
                 dimension.errored for dimension in self.dimensions
@@ -131,28 +167,28 @@ class Explore(LookMlObject):
             return None
 
     @errored.setter
-    def errored(self, value):
+    def errored(self, value: bool) -> None:
         raise AttributeError(
             "Cannot assign to 'errored' property of an Explore instance. "
             "For an explore to be considered errored, it must have a ValidationError "
             "in its 'errors' attribute or contain dimensions in an errored state."
         )
 
-    def get_errored_dimensions(self):
+    def get_errored_dimensions(self) -> Generator[Dimension, None, None]:
         for dimension in self.dimensions:
             if dimension.errored:
                 yield dimension
 
     @classmethod
-    def from_json(cls, json_dict, model_name):
+    def from_json(cls, json_dict: Dict[str, Any], model_name: str) -> "Explore":
         name = json_dict["name"]
         return cls(name, model_name)
 
-    def add_dimension(self, dimension: Dimension):
+    def add_dimension(self, dimension: Dimension) -> None:
         self.dimensions.append(dimension)
 
     @property
-    def number_of_errors(self):
+    def number_of_errors(self) -> int:
         if self.errored:
             if self.errors:
                 errors = len(self.errors)
@@ -167,6 +203,29 @@ class Explore(LookMlObject):
             return 0
 
 
+@dataclass(eq=True, frozen=True)
+class CompiledSql:
+    model_name: str
+    explore_name: str
+    sql: str
+    dimension_name: Optional[str] = None
+
+    @classmethod
+    def from_explore(cls, explore: Explore, sql: str) -> "CompiledSql":
+        return CompiledSql(
+            model_name=explore.model_name, explore_name=explore.name, sql=sql
+        )
+
+    @classmethod
+    def from_dimension(cls, dimension: Dimension, sql: str) -> "CompiledSql":
+        return CompiledSql(
+            model_name=dimension.model_name,
+            explore_name=dimension.explore_name,
+            dimension_name=dimension.name,
+            sql=sql,
+        )
+
+
 class Model(LookMlObject):
     def __init__(self, name: str, project_name: str, explores: List[Explore]):
         self.name = name
@@ -174,10 +233,10 @@ class Model(LookMlObject):
         self.explores = explores
         self.errors: List[ValidationError] = []
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}(name={self.name}, explores={self.explores})"
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, Model):
             return NotImplemented
 
@@ -188,7 +247,7 @@ class Model(LookMlObject):
         )
 
     @property
-    def errored(self):
+    def errored(self) -> Optional[bool]:
         if self.queried:
             return bool(self.errors) or any(
                 explore.errored for explore in self.explores
@@ -197,7 +256,7 @@ class Model(LookMlObject):
             return None
 
     @errored.setter
-    def errored(self, value: bool):
+    def errored(self, value: bool) -> None:
         if not isinstance(value, bool):
             raise TypeError("Value for errored must be boolean.")
         if not self.explores:
@@ -208,11 +267,11 @@ class Model(LookMlObject):
             explore.errored = value
 
     @property
-    def queried(self):
+    def queried(self) -> bool:
         return any(explore.queried for explore in self.explores)
 
     @queried.setter
-    def queried(self, value: bool):
+    def queried(self, value: bool) -> None:
         if not isinstance(value, bool):
             raise TypeError("Value for queried must be boolean.")
         for explore in self.explores:
@@ -223,13 +282,13 @@ class Model(LookMlObject):
             (explore for explore in self.explores if explore.name == name), None
         )
 
-    def get_errored_explores(self):
+    def get_errored_explores(self) -> Generator[Explore, None, None]:
         for explore in self.explores:
             if explore.errored:
                 yield explore
 
     @classmethod
-    def from_json(cls, json_dict):
+    def from_json(cls, json_dict: Dict[str, Any]) -> "Model":
         name = json_dict["name"]
         project = json_dict["project_name"]
         explores = [
@@ -238,18 +297,18 @@ class Model(LookMlObject):
         return cls(name, project, explores)
 
     @property
-    def number_of_errors(self):
+    def number_of_errors(self) -> int:
         return len(self.errors) + sum(
             [explore.number_of_errors for explore in self.explores if explore.errored]
         )
 
 
 class Project(LookMlObject):
-    def __init__(self, name, models: Sequence[Model]):
+    def __init__(self, name: str, models: Sequence[Model]) -> None:
         self.name = name
         self.models = models
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, Project):
             return NotImplemented
 
@@ -286,14 +345,14 @@ class Project(LookMlObject):
                     yield dimension
 
     @property
-    def errored(self):
+    def errored(self) -> Optional[bool]:
         if self.queried:
             return any(model.errored for model in self.models)
         else:
             return None
 
     @errored.setter
-    def errored(self, value: bool):
+    def errored(self, value: bool) -> None:
         if not isinstance(value, bool):
             raise TypeError("Value for errored must be boolean.")
         if not self.models:
@@ -304,11 +363,11 @@ class Project(LookMlObject):
             model.errored = value
 
     @property
-    def queried(self):
+    def queried(self) -> bool:
         return any(model.queried for model in self.models)
 
     @queried.setter
-    def queried(self, value: bool):
+    def queried(self, value: bool) -> None:
         if not isinstance(value, bool):
             raise TypeError("Value for queried must be boolean.")
         for model in self.models:
@@ -356,17 +415,26 @@ class Project(LookMlObject):
 
             for explore in model.explores:
                 if filters is not None and not is_selected(
-                    model.name, explore.name, filters
+                    model.name,
+                    explore.name,
+                    filters,  # pyright: ignore[reportGeneralTypeIssues]
                 ):
                     continue
-                status = "passed"
+
+                test: Dict[str, Any] = {
+                    "model": model.name,
+                    "explore": explore.name,
+                    "status": "passed",  # To be overwritten if needed
+                }
+
                 if explore.skipped:
-                    status = "skipped"
+                    test["status"] = "skipped"
+                    test["skip_reason"] = explore.skipped.value
                 elif explore.errored and validator != "sql":
-                    status = "failed"
+                    test["status"] = "failed"
                     errors.extend([e.to_dict() for e in explore.errors])
                 elif explore.errored and fail_fast is True:
-                    status = "failed"
+                    test["status"] = "failed"
                     errors.append(explore.errors[0].to_dict())
                 elif explore.errored:
                     dimension_errors = [e for d in explore.dimensions for e in d.errors]
@@ -374,17 +442,13 @@ class Project(LookMlObject):
                     # errors, return those instead. Skip anything marked as ignored.
                     relevant_errors = [
                         e.to_dict()
-                        for e in (dimension_errors or explore.errors)
+                        for e in dimension_errors + explore.errors
                         if not e.ignore
                     ]
                     if relevant_errors:
-                        status = "failed"
+                        test["status"] = "failed"
                         errors.extend(relevant_errors)
-                test: Dict[str, Any] = {
-                    "model": model.name,
-                    "explore": explore.name,
-                    "status": status,
-                }
+
                 if explore.successes:
                     successes.extend([success for success in explore.successes])
 
@@ -399,32 +463,44 @@ class Project(LookMlObject):
             "successes": successes,
         }
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}(name={self.name}, models={self.models})"
 
     @property
-    def number_of_errors(self):
+    def number_of_errors(self) -> int:
         return sum([model.number_of_errors for model in self.models if model.errored])
 
 
-def build_dimensions(
+async def build_explore_dimensions(
     client: LookerClient,
-    model_name: str,
-    explore_name: str,
+    explore: Explore,
     ignore_hidden_fields: bool = False,
-) -> List[Dimension]:
+) -> None:
     """Creates Dimension objects for all dimensions in a given explore."""
-    dimensions_json = client.get_lookml_dimensions(model_name, explore_name)
+    dimensions_json = await client.get_lookml_dimensions(
+        explore.model_name, explore.name
+    )
+
     dimensions: List[Dimension] = []
     for dimension_json in dimensions_json:
-        dimension = Dimension.from_json(dimension_json, model_name, explore_name)
-        dimension.url = client.base_url + dimension.url
+        dimension: Dimension = Dimension.from_json(
+            dimension_json, explore.model_name, explore.name
+        )
+        if dimension.url is not None:
+            dimension.url = client.base_url + dimension.url
         if not dimension.ignore and not (dimension.is_hidden and ignore_hidden_fields):
             dimensions.append(dimension)
-    return dimensions
+
+    explore.dimensions = dimensions
+    if len(explore.dimensions) == 0:
+        logger.warning(
+            f"Warning: Explore '{explore.name}' does not have any non-ignored "
+            "dimensions and will not be validated."
+        )
+        explore.skipped = SkipReason.NO_DIMENSIONS
 
 
-def build_project(
+async def build_project(
     client: LookerClient,
     name: str,
     filters: Optional[List[str]] = None,
@@ -438,7 +514,7 @@ def build_project(
 
     models = []
     fields = ["name", "project_name", "explores"]
-    for lookmlmodel in client.get_lookml_models(fields=fields):
+    for lookmlmodel in await client.get_lookml_models(fields=fields):
         model = Model.from_json(lookmlmodel)
         if model.project_name == name:
             models.append(model)
@@ -456,19 +532,22 @@ def build_project(
 
     # Prune to selected explores for non-content validators
     if not include_all_explores:
+        tasks: List[asyncio.Task[Any]] = []
         for model in models:
             model.explores = [
                 explore
                 for explore in model.explores
                 if is_selected(model.name, explore.name, filters)
             ]
+            if include_dimensions:
+                for explore in model.explores:
+                    task = asyncio.create_task(
+                        build_explore_dimensions(client, explore, ignore_hidden_fields),
+                        name=f"build_explore_dimensions_{explore.name}",
+                    )
+                    tasks.append(task)
 
-    if include_dimensions:
-        for model in models:
-            for explore in model.explores:
-                explore.dimensions = build_dimensions(
-                    client, model.name, explore.name, ignore_hidden_fields
-                )
+        await asyncio.gather(*tasks)
 
     # Include empty models when including all explores
     if include_all_explores:
