@@ -1,27 +1,28 @@
 import asyncio
-import re
-from spectacles.exceptions import LookerApiError, SqlError
-from typing import Dict, List, Optional, Tuple, Set
-from dataclasses import dataclass
 import itertools
-from spectacles.client import LookerClient
+import re
+from dataclasses import dataclass
+from typing import Any, Dict, List, Optional, Set, Tuple
+
+from spectacles.client import LOOKML_VALIDATION_TIMEOUT, LookerClient
+from spectacles.exceptions import LookerApiError, SpectaclesException, SqlError
+from spectacles.logger import GLOBAL_LOGGER as logger
+from spectacles.lookml import CompiledSql, Explore, build_project
+from spectacles.models import JsonDict, SkipReason
+from spectacles.printer import print_header
+from spectacles.utils import time_hash
 from spectacles.validators import (
-    SqlValidator,
-    DataTestValidator,
     ContentValidator,
+    DataTestValidator,
     LookMLValidator,
+    SqlValidator,
 )
-from spectacles.types import JsonDict
+from spectacles.validators.data_test import DATA_TEST_CONCURRENCY
 from spectacles.validators.sql import (
     DEFAULT_CHUNK_SIZE,
     DEFAULT_QUERY_CONCURRENCY,
     DEFAULT_RUNTIME_THRESHOLD,
 )
-from spectacles.exceptions import SpectaclesException
-from spectacles.utils import time_hash
-from spectacles.lookml import CompiledSql, build_project, Explore
-from spectacles.logger import GLOBAL_LOGGER as logger
-from spectacles.printer import print_header
 
 
 @dataclass
@@ -43,6 +44,7 @@ class LookerBranchManager:
         client: LookerClient,
         project: str,
         remote_reset: bool = False,
+        use_personal_branch: bool = False,
         pin_imports: Optional[Dict[str, str]] = None,
         skip_imports: Optional[List[str]] = None,
     ):
@@ -57,10 +59,14 @@ class LookerBranchManager:
         self.commit: Optional[str] = None
         self.branch: Optional[str] = None
         self.is_temp_branch: bool = False
+        self.use_personal_branch: bool = use_personal_branch
+        self.personal_branch: Optional[str] = None
         self.import_managers: List[LookerBranchManager] = []
         self.skip_imports: List[str] = [] if skip_imports is None else skip_imports
 
-    def __call__(self, ref: Optional[str] = None, ephemeral: Optional[bool] = None):
+    def __call__(
+        self, ref: Optional[str] = None, ephemeral: Optional[bool] = None
+    ) -> "LookerBranchManager":
         logger.debug("")
         logger.debug(
             f"Setting Git state for project '{self.project}' "
@@ -77,7 +83,7 @@ class LookerBranchManager:
             self.branch = ref
 
         if ephemeral is None:
-            if self.commit:
+            if self.commit or self.pin_imports:
                 self.ephemeral = True
             else:
                 self.ephemeral = False
@@ -95,7 +101,7 @@ class LookerBranchManager:
         self.import_managers = []
         return self
 
-    async def __aenter__(self):
+    async def __aenter__(self) -> None:
         logger.indent(1)
 
         state: ProjectState = await self.get_project_state()
@@ -105,7 +111,11 @@ class LookerBranchManager:
         if self.branch:
             await self.update_workspace("dev")
             if self.ephemeral:
-                self.branch = await self.checkout_temp_branch("origin/" + self.branch)
+                new_branch = await self.checkout_ephemeral_branch(
+                    "origin/" + self.branch
+                )
+                if not self.use_personal_branch:
+                    self.branch = new_branch
             else:
                 await self.client.checkout_branch(self.project, self.branch)
                 if self.remote_reset:
@@ -113,7 +123,7 @@ class LookerBranchManager:
         # A commit was passed, so we non-destructively create a temporary branch we can
         # hard reset to the commit.
         elif self.commit:
-            self.branch = await self.checkout_temp_branch(self.commit)
+            self.branch = await self.checkout_ephemeral_branch(self.commit)
         # Neither branch nor commit were passed, so go to production.
         else:
             if self.init_state.workspace == "production":
@@ -124,7 +134,7 @@ class LookerBranchManager:
             self.branch = prod_state.branch
             self.commit = prod_state.commit
             if self.ephemeral:
-                self.branch = await self.checkout_temp_branch(prod_state.commit)
+                self.branch = await self.checkout_ephemeral_branch(prod_state.commit)
 
         logger.debug(
             f"Set project '{self.project}' to branch '{self.branch}' @ "
@@ -160,6 +170,7 @@ class LookerBranchManager:
                         project,
                         pin_imports=self.pin_imports,
                         skip_imports=self.skip_imports,
+                        use_personal_branch=self.use_personal_branch,
                     )
                     await manager(ref=import_ref, ephemeral=True).__aenter__()
                     self.import_managers.append(manager)
@@ -172,7 +183,7 @@ class LookerBranchManager:
         logger.indent(-1)
         logger.debug("")
 
-    async def __aexit__(self, *args):
+    async def __aexit__(self, *args: Any) -> None:
         logger.debug("")
         logger.debug(f"Cleaning up Git state in '{self.project}'")
         logger.indent(1)
@@ -196,6 +207,8 @@ class LookerBranchManager:
         for manager in self.import_managers:
             await manager.__aexit__()
 
+        self.skip_imports = []
+
         if self.init_state.workspace == "production":
             await self.update_workspace("production")
         else:
@@ -223,7 +236,7 @@ class LookerBranchManager:
         else:
             return self.branch
 
-    async def update_workspace(self, workspace: str):
+    async def update_workspace(self, workspace: str) -> None:
         if workspace not in ("dev", "production"):
             raise ValueError("Workspace can only be set to 'dev' or 'production'")
         if self.workspace != workspace:
@@ -245,6 +258,27 @@ class LookerBranchManager:
         else:
             return [p["name"] for p in manifest["imports"] if not p["is_remote"]]
 
+    async def checkout_personal_branch(self, ref: str) -> str:
+        """Updates the user's personal branch to the git ref."""
+        await self.update_workspace("dev")
+        if not self.personal_branch:
+            self.personal_branch = await self.get_personal_branch()
+        await self.client.checkout_branch(self.project, self.personal_branch)
+        await self.client.reset_to_remote(self.project)
+        await self.client.hard_reset_branch(self.project, self.personal_branch, ref)
+        return self.personal_branch
+
+    async def get_personal_branch(self) -> str:
+        """Finds the name of the user's personal branch."""
+        branches = await self.client.get_all_branches(self.project)
+        for branch in branches:
+            if branch["personal"] and not branch["readonly"]:
+                return str(branch["name"])
+        raise ValueError(
+            f"Personal branch not found for client ID {self.client.client_id} "
+            f"in project '{self.project}'"
+        )
+
     async def checkout_temp_branch(self, ref: str) -> str:
         """Creates a temporary branch off a commit or off production."""
         # Save the dev mode state so we have somewhere to delete the temp branch
@@ -262,6 +296,14 @@ class LookerBranchManager:
         self.is_temp_branch = True
         return name
 
+    async def checkout_ephemeral_branch(self, ref: str) -> str:
+        """Either check out temp or personal branch and hard-reset."""
+        if self.use_personal_branch:
+            branch = await self.checkout_personal_branch(ref)
+        else:
+            branch = await self.checkout_temp_branch(ref)
+        return branch
+
 
 class Runner:
     """Runs validations and returns JSON-style dictionaries with validation results.
@@ -286,11 +328,16 @@ class Runner:
         project: str,
         remote_reset: bool = False,
         pin_imports: Optional[Dict[str, str]] = None,
+        use_personal_branch: bool = False,
     ):
         self.project = project
         self.client = client
         self.branch_manager = LookerBranchManager(
-            client, project, remote_reset, pin_imports or {}
+            client,
+            project,
+            remote_reset,
+            pin_imports=pin_imports or {},
+            use_personal_branch=use_personal_branch,
         )
 
     async def validate_sql(
@@ -305,9 +352,20 @@ class Runner:
         runtime_threshold: int = DEFAULT_RUNTIME_THRESHOLD,
         chunk_size: int = DEFAULT_CHUNK_SIZE,
         ignore_hidden_fields: bool = False,
+        result_format: str = "json_bi",
     ) -> JsonDict:
         if filters is None:
             filters = ["*/*"]
+        if profile and result_format == "json_bi":
+            raise SpectaclesException(
+                name="profiling-incompatible-with-json-bi",
+                title="Profiling incompatible with json_bi result format",
+                detail=(
+                    "Profiling query runtimes is incompatible with the "
+                    "json_bi result format. Please use the legacy json_detail "
+                    "result format if enabling profiling (--use-legacy-result-format)."
+                ),
+            )
         validator = SqlValidator(self.client, concurrency, runtime_threshold)
         ephemeral = True if incremental else None
         # Create explore-level tests for the desired ref
@@ -347,22 +405,15 @@ class Runner:
                         ),
                     )
                 logger.debug("Compiling SQL for Explores at the target ref")
-                target_project = await build_project(
-                    self.client,
-                    name=self.project,
-                    filters=filters,
-                    include_dimensions=True,
-                    ignore_hidden_fields=ignore_hidden_fields,
-                )
                 target_explores: Set[CompiledSql] = set()
-                if incremental:
-                    compiled_explores = await asyncio.gather(
-                        *(
-                            validator.compile_explore(explore)
-                            for explore in target_project.iter_explores()
-                        )
+
+                compiled_explores = await asyncio.gather(
+                    *(
+                        validator.compile_explore(explore)
+                        for explore in project.iter_explores()
                     )
-                    target_explores = set(compiled_explores)
+                )
+                target_explores = set(compiled_explores)
 
             # Determine which explore tests are identical between target and base
             # Iterate instead of set operations so we have control of which test, and
@@ -379,7 +430,7 @@ class Runner:
                     )
                 if compiled in target_explores:
                     # Mark explores with the same compiled SQL (test) as skipped
-                    explore.skipped = True
+                    explore.skipped = SkipReason.UNMODIFIED
                 else:
                     # Test explores with unique SQL for base ref
                     explores += (explore,)
@@ -388,17 +439,21 @@ class Runner:
         else:
             explores = tuple(project.iter_explores())
 
-        explore_count = len(explores)
+        n_tested_explores = len([e for e in explores if not e.skipped])
+        n_total_explores = len(base_explores or explores)
         print_header(
-            f"Testing {explore_count} "
-            f"{'explore' if explore_count == 1 else 'explores'} "
+            f"Testing {n_tested_explores}/{n_total_explores} explores "
             + ("[fail fast] " if fail_fast else "")
             + f"[concurrency = {concurrency}]"
         )
 
         async with self.branch_manager(ref=ref):
             await validator.search(
-                explores, fail_fast, chunk_size, profile=profile if fail_fast else False
+                explores,
+                fail_fast,
+                chunk_size,
+                profile=profile,
+                result_format=result_format,
             )
 
         # Create dimension tests for the desired ref when explores errored
@@ -442,6 +497,7 @@ class Runner:
         self,
         ref: Optional[str] = None,
         filters: Optional[List[str]] = None,
+        concurrency: int = DATA_TEST_CONCURRENCY,
     ) -> JsonDict:
         if filters is None:
             filters = ["*/*"]
@@ -460,16 +516,21 @@ class Runner:
                 f"{'explore' if explore_count == 1 else 'explores'}"
             )
             tests = await validator.get_tests(project)
-            await validator.validate(tests)
+            await validator.validate(tests, concurrency)
 
         results = project.get_results(validator="data_test")
         return results
 
-    async def validate_lookml(self, ref: Optional[str], severity: str) -> JsonDict:
+    async def validate_lookml(
+        self,
+        ref: Optional[str],
+        severity: str,
+        timeout: int = LOOKML_VALIDATION_TIMEOUT,
+    ) -> JsonDict:
         async with self.branch_manager(ref=ref):
             validator = LookMLValidator(self.client)
             print_header(f"Validating LookML in project {self.project} [{severity}]")
-            results = await validator.validate(self.project, severity)
+            results = await validator.validate(self.project, severity, timeout)
         return results
 
     async def validate_content(
@@ -555,7 +616,7 @@ class Runner:
             test = dict(model=error["model"], explore=error["explore"], status=status)
             tests.append(test)
 
-        def key_by(x):
+        def key_by(x: dict[str, Any]) -> Tuple[str, str]:
             return (x["model"], x["explore"])
 
         if tests:
