@@ -1,14 +1,18 @@
-from typing import List, Callable, Optional, Dict, Any, Iterable
-from urllib import parse
-from spectacles.logger import GLOBAL_LOGGER as logger
-import functools
-import requests
-import timeit
+from __future__ import annotations
+
+import asyncio
 import hashlib
 import time
+from typing import Any, Callable, Coroutine, Dict, Iterable, List, Optional, Tuple
+from urllib import parse
+
+import httpx
+
+from spectacles.logger import GLOBAL_LOGGER as logger
+from spectacles.models import T
 
 
-def compose_url(base_url: str, path: List, params: Dict = {}) -> str:
+def compose_url(base_url: str, path: List[str], params: Dict[str, Any] = {}) -> str:
     if not isinstance(path, list):
         raise TypeError("URL path must be a list")
     path_parts = [base_url] + path
@@ -25,16 +29,16 @@ def compose_url(base_url: str, path: List, params: Dict = {}) -> str:
     return url
 
 
-def details_from_http_error(response: requests.Response) -> Optional[Dict[str, Any]]:
+def details_from_http_error(response: httpx.Response) -> Optional[Dict[str, Any]]:
     try:
         details = response.json()
     # Requests raises a ValueError if the response is invalid JSON
     except ValueError:
         details = None
-    return details
+    return details  # type: ignore[no-any-return]
 
 
-def human_readable(elapsed: int):
+def human_readable(elapsed: float) -> str:
     minutes, seconds = divmod(elapsed, 60)
     num_mins = f"{minutes:.0f} minute{'s' if minutes > 1 else ''}"
     num_secs = f"{seconds:.0f} second{'s' if round(seconds) != 1 else ''}"
@@ -43,7 +47,7 @@ def human_readable(elapsed: int):
     return f"{num_mins if minutes else ''}{separator}{num_secs if seconds else ''}"
 
 
-def get_detail(fn_name: str):
+def get_detail(fn_name: str) -> str:
     detail_map = {
         "run_sql": "SQL ",
         "run_assert": "data test ",
@@ -52,16 +56,16 @@ def get_detail(fn_name: str):
     return detail_map.get(fn_name, "")
 
 
-def log_duration(fn: Callable):
-    functools.wraps(fn)
-
-    def timed_function(*args, **kwargs):
-        start_time = timeit.default_timer()
+def log_duration(
+    fn: Callable[..., Coroutine[Any, Any, Any]]
+) -> Callable[..., Coroutine[Any, Any, Any]]:
+    async def timed_function(*args: Any, **kwargs: Any) -> Any:
+        start_time = time.time()
         try:
-            result = fn(*args, **kwargs)
+            result = await fn(*args, **kwargs)
         finally:
-            elapsed = timeit.default_timer() - start_time
-            elapsed_str = human_readable(elapsed)
+            elapsed = time.time() - start_time
+            elapsed_str = human_readable(int(elapsed))
             message_detail = get_detail(fn.__name__)
 
             logger.info(f"Completed {message_detail}validation in {elapsed_str}.\n")
@@ -76,7 +80,28 @@ def time_hash() -> str:
     return hash.hexdigest()[:10]
 
 
-def chunks(to_chunk: list, size: int) -> Iterable:
+def chunks(to_chunk: list[T], size: int) -> Iterable[list[T]]:
     """Yield successive n-sized chunks from the list."""
     for i in range(0, len(to_chunk), size):
         yield to_chunk[i : i + size]
+
+
+def consume_queue(
+    queue: asyncio.Queue[T], limit: Optional[int] = None
+) -> Tuple[T, ...]:
+    """Purge an async queue of all its contents, up to a limit, and return them."""
+    count = 0
+    contents: tuple[T, ...] = tuple()
+    while not queue.empty() and (limit is None or count <= limit):
+        contents += (queue.get_nowait(),)
+        count += 1
+    return contents
+
+
+def halt_queue(queue: asyncio.Queue[Any]) -> None:
+    """Inform a queue that all tasks are finished, unblocking any Queue.join calls."""
+    while True:
+        try:
+            queue.task_done()
+        except ValueError:  # ValueError raised when no unfinished tasks remain
+            break
