@@ -1,8 +1,14 @@
+import asyncio
 from dataclasses import dataclass
 from typing import List, Optional
+
 from spectacles.client import LookerClient
+from spectacles.exceptions import DataTestError, SpectaclesException
 from spectacles.lookml import Explore, Project
-from spectacles.exceptions import SpectaclesException, DataTestError
+
+DATA_TEST_CONCURRENCY = (
+    15  # This is the per-user query limit in Looker for most instances
+)
 
 
 @dataclass
@@ -16,7 +22,7 @@ class DataTest:
     line: int
     passed: Optional[bool] = None
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         try:
             self.file_path = self.file.split("/", 1)[1]
         except IndexError:
@@ -27,7 +33,7 @@ class DataTest:
             )
 
     @property
-    def explore_url(self):
+    def explore_url(self) -> str:
         return (
             f"{self.base_url}/explore/{self.explore.model_name}"
             f"/{self.explore.name}?{self.query_url_params}"
@@ -53,8 +59,8 @@ class DataTestValidator:
     def __init__(self, client: LookerClient):
         self.client = client
 
-    def get_tests(self, project: Project) -> List[DataTest]:
-        all_tests = self.client.all_lookml_tests(project.name)
+    async def get_tests(self, project: Project) -> List[DataTest]:
+        all_tests = await self.client.all_lookml_tests(project.name)
 
         # Filter the list of tests to those that are selected
         selected_tests: List[DataTest] = []
@@ -90,40 +96,46 @@ class DataTestValidator:
 
         return selected_tests
 
-    def validate(self, tests: List[DataTest]) -> List[DataTestError]:
+    async def validate(
+        self, tests: List[DataTest], concurrency: int = DATA_TEST_CONCURRENCY
+    ) -> List[DataTestError]:
         data_test_errors: List[DataTestError] = []
-        for test in tests:
-            results = self.client.run_lookml_test(
-                test.project_name, model=test.explore.model_name, test=test.name
-            )
-            test.explore.queried = True
-            result = results[0]  # For a single test, list with length 1
+        query_slot = asyncio.Semaphore(concurrency)
 
-            if result["success"]:
-                test.passed = True
-                test.explore.successes.append(
-                    {
-                        "model": test.explore.model_name,
-                        "explore": test.explore.name,
-                        "metadata": {
-                            "test_name": result["test_name"],
-                            "lookml_url": test.lookml_url,
-                            "explore_url": test.explore_url,
-                        },
-                    }
+        async def run_test(test: DataTest, query_slot: asyncio.Semaphore) -> None:
+            async with query_slot:
+                results = await self.client.run_lookml_test(
+                    test.project_name, model=test.explore.model_name, test=test.name
                 )
-            else:
-                test.passed = False
-                for error in result["errors"]:
-                    error = DataTestError(
-                        model=error["model_id"],
-                        explore=error["explore"],
-                        message=error["message"],
-                        test_name=result["test_name"],
-                        lookml_url=test.lookml_url,
-                        explore_url=test.explore_url,
-                    )
-                    data_test_errors.append(error)
-                    test.explore.errors.append(error)
+                test.explore.queried = True
+                result = results[0]  # For a single test, list with length 1
 
+                if result["success"]:
+                    test.passed = True
+                    test.explore.successes.append(
+                        {
+                            "model": test.explore.model_name,
+                            "explore": test.explore.name,
+                            "metadata": {
+                                "test_name": result["test_name"],
+                                "lookml_url": test.lookml_url,
+                                "explore_url": test.explore_url,
+                            },
+                        }
+                    )
+                else:
+                    test.passed = False
+                    for error in result["errors"]:
+                        error = DataTestError(
+                            model=error["model_id"],
+                            explore=error["explore"],
+                            message=error["message"],
+                            test_name=result["test_name"],
+                            lookml_url=test.lookml_url,
+                            explore_url=test.explore_url,
+                        )
+                        data_test_errors.append(error)
+                        test.explore.errors.append(error)
+
+        await asyncio.gather(*(run_test(test, query_slot) for test in tests))
         return data_test_errors
