@@ -1,22 +1,33 @@
 from __future__ import annotations
+
 import asyncio
+import time
 from dataclasses import dataclass
-from tabulate import tabulate
-from typing import List, Optional, Tuple, Iterator, Union, cast
+from typing import Iterator, List, Optional, Tuple
+
 import pydantic
+from tabulate import tabulate
+
 from spectacles.client import LookerClient
-from spectacles.lookml import CompiledSql, Dimension, Explore
 from spectacles.exceptions import SpectaclesException, SqlError
 from spectacles.logger import GLOBAL_LOGGER as logger
+from spectacles.lookml import CompiledSql, Dimension, Explore
+from spectacles.models import (
+    CompletedQueryResult,
+    ErrorQueryResult,
+    InterruptedQueryResult,
+    QueryResult,
+)
 from spectacles.printer import print_header
 from spectacles.utils import consume_queue, halt_queue
-from spectacles.types import QueryResult, CompletedQueryResult, ErrorQueryResult
 
 QUERY_TASK_LIMIT = 250
 DEFAULT_CHUNK_SIZE = 500
 DEFAULT_QUERY_CONCURRENCY = 10
 DEFAULT_RUNTIME_THRESHOLD = 5
-ProfilerTableRow = Tuple[str, float, str, str]
+EXPIRED_QUERY_WAIT_TIME = 300
+EXPIRED_RETRY_LIMIT = 1
+ProfilerTableRow = Tuple[str, str, float, str, str]
 
 
 @dataclass
@@ -27,6 +38,8 @@ class Query:
     explore_url: str | None = None
     errored: bool | None = None
     runtime: float | None = None
+    expired_at: float | None = None
+    expired_retries: int = 0
 
     def __post_init__(self) -> None:
         # Confirm that all dimensions are from the Explore associated here
@@ -62,7 +75,13 @@ class Query:
                 "Query.explore_url cannot be None, "
                 "run Query.create to get an explore URL"
             )
-        return (self.explore.name, self.runtime, self.query_id, self.explore_url)
+        return (
+            self.explore.name,
+            self.dimensions[0].name if len(self.dimensions) == 1 else "*",
+            self.runtime,
+            self.query_id,
+            self.explore_url,
+        )
 
 
 def print_profile_results(queries: List[Query], runtime_threshold: int) -> None:
@@ -79,6 +98,7 @@ def print_profile_results(queries: List[Query], runtime_threshold: int) -> None:
             [query.to_profiler_format() for query in queries_by_runtime],
             headers=[
                 "Explore",
+                "Dimension(s)",
                 "Runtime (s)",
                 "Query ID",
                 "Explore From Here",
@@ -122,30 +142,41 @@ class SqlValidator:
         self._long_running_queries: List[Query] = []
 
     async def compile_explore(self, explore: Explore) -> CompiledSql:
-        if not explore.dimensions:
-            raise AttributeError(
-                f"Explore '{explore.name}' is missing dimensions, "
-                "meaning this query won't have fields and will error. "
-                "Often this happens because you didn't include dimensions "
-                "when you built the project."
+        if explore.skipped:
+            sql = ""
+        else:
+            dimensions = [dimension.name for dimension in explore.dimensions]
+            query_body = {
+                "model": explore.model_name,
+                "view": explore.name,
+                "fields": dimensions,
+                "limit": 0,
+                "filter_expression": "1=2",
+            }
+            sql = await self.client.run_inline_query(
+                query_body=query_body,
+                result_format="sql",
+                model=explore.model_name,
+                explore=explore.name,
             )
-        dimensions = [dimension.name for dimension in explore.dimensions]
-        # Create a query that includes all dimensions
-        query = await self.client.create_query(
-            explore.model_name, explore.name, dimensions, fields=["id"]
-        )
-        sql = await self.client.run_query(query["id"])
+
         return CompiledSql.from_explore(explore, sql)
 
     async def compile_dimension(self, dimension: Dimension) -> CompiledSql:
-        # Create a query for the dimension
-        query = await self.client.create_query(
-            dimension.model_name,
-            dimension.explore_name,
-            [dimension.name],
-            fields=["id"],
+        query_body = {
+            "model": dimension.model_name,
+            "view": dimension.explore_name,
+            "fields": [dimension.name],
+            "limit": 0,
+            "filter_expression": "1=2",
+        }
+        sql = await self.client.run_inline_query(
+            query_body=query_body,
+            result_format="sql",
+            model=dimension.model_name,
+            explore=dimension.explore_name,
+            dimension=dimension.name,
         )
-        sql = await self.client.run_query(query["id"])
         return CompiledSql.from_dimension(dimension, sql)
 
     async def search(
@@ -154,6 +185,7 @@ class SqlValidator:
         fail_fast: bool,
         chunk_size: int = DEFAULT_CHUNK_SIZE,
         profile: bool = False,
+        result_format: str = "json_bi",
     ) -> None:
         queries_to_run: asyncio.Queue[Optional[Query]] = asyncio.Queue()
         running_queries: asyncio.Queue[str] = asyncio.Queue()
@@ -161,7 +193,12 @@ class SqlValidator:
 
         workers = (
             asyncio.create_task(
-                self._run_query(queries_to_run, running_queries, query_slot),
+                self._run_query(
+                    queries_to_run,
+                    running_queries,
+                    query_slot,
+                    result_format,
+                ),
                 name="run_query",
             ),
             asyncio.create_task(
@@ -176,11 +213,8 @@ class SqlValidator:
             for explore in explores:
                 # Sorting makes it more likely to prune the tree faster in binsearch
                 dimensions = tuple(sorted(explore.dimensions))
-                if len(dimensions) == 0:
-                    logger.warning(
-                        f"Warning: Explore '{explore.name}' does not have any non-ignored "
-                        "dimensions and will not be validated."
-                    )
+                if explore.skipped:
+                    continue
                 elif len(dimensions) <= chunk_size:
                     queries_to_run.put_nowait(Query(explore, dimensions))
                 else:
@@ -235,6 +269,7 @@ class SqlValidator:
         queries_to_run: asyncio.Queue[Optional[Query]],
         running_queries: asyncio.Queue[str],
         query_slot: asyncio.Semaphore,
+        result_format: str = "json_bi",
     ) -> None:
         try:
             # End execution if a sentinel is received from the queue
@@ -255,7 +290,9 @@ class SqlValidator:
                         "Query.query_id cannot be None, "
                         "run Query.create to get a query ID"
                     )
-                task_id = await self.client.create_query_task(query.query_id)
+                task_id = await self.client.create_query_task(
+                    query.query_id, result_format
+                )
                 self._task_to_query[task_id] = query
                 running_queries.put_nowait(task_id)
 
@@ -292,7 +329,7 @@ class SqlValidator:
                 raw = await self.client.get_query_task_multi_results(task_ids)
                 for task_id, result in raw.items():
                     try:
-                        query_result = QueryResult.parse_obj(result).__root__
+                        query_result = QueryResult.model_validate(result).root
                     except pydantic.ValidationError as validation_error:
                         logger.debug(
                             f"Unable to parse unexpected Looker API response format: {result}"
@@ -310,12 +347,15 @@ class SqlValidator:
                     )
 
                     # Append long-running queries for the profiler
-                    if query_result.status in ("complete", "error"):
-                        query_result = cast(
-                            Union[CompletedQueryResult, ErrorQueryResult], query_result
-                        )
+                    if isinstance(
+                        query_result, (CompletedQueryResult, ErrorQueryResult)
+                    ):
                         query = self._task_to_query[task_id]
-                        if query_result.runtime > self.runtime_threshold:
+                        query.runtime = query_result.runtime
+                        if (
+                            query_result.runtime
+                            and query_result.runtime > self.runtime_threshold
+                        ):
                             self._long_running_queries.append(query)
                         if query_result.status == "complete":
                             query_slot.release()
@@ -385,6 +425,80 @@ class SqlValidator:
 
                             # Indicate there are no more queries or subqueries to run
                             queries_to_run.task_done()
+
+                    elif (
+                        isinstance(query_result, InterruptedQueryResult)
+                        and query_result.status == "killed"
+                    ):
+                        query = self._task_to_query[task_id]
+                        query.errored = True
+                        explore = query.explore
+                        explore.queried = True
+                        explore.errors.append(
+                            SqlError(
+                                model=explore.model_name,
+                                explore=explore.name,
+                                dimension=None,
+                                sql="",
+                                message=(
+                                    "Couldn't finish testing "
+                                    f"{explore.model_name}.{explore.name} "
+                                    "because the test query was killed "
+                                    "in the database. "
+                                ),
+                                explore_url=query.explore_url,
+                            )
+                        )
+                        query_slot.release()
+                        queries_to_run.task_done()
+
+                    elif (
+                        isinstance(query_result, InterruptedQueryResult)
+                        and query_result.status == "expired"
+                    ):
+                        query = self._task_to_query[task_id]
+                        query.expired_at = query.expired_at or time.time()
+                        expired_for = time.time() - query.expired_at
+                        if expired_for > EXPIRED_QUERY_WAIT_TIME:
+                            # Stop waiting for query, decide if we should retry
+                            if query.expired_retries < EXPIRED_RETRY_LIMIT:
+                                logger.debug(
+                                    f"Query task {task_id} expired for "
+                                    f"over {EXPIRED_QUERY_WAIT_TIME} seconds. "
+                                    "Creating a new query task to try again."
+                                )
+                                query.expired_at = None
+                                query.expired_retries += 1
+                                await queries_to_run.put(query)
+                            else:
+                                logger.debug(
+                                    f"Query task {task_id} keeps expiring, "
+                                    f"even after {EXPIRED_RETRY_LIMIT + 1} tries. "
+                                    "Giving up on it."
+                                )
+                                query.errored = True
+                                explore = query.explore
+                                explore.queried = True  # Not exactly, but close enough
+                                explore.errors.append(
+                                    SqlError(
+                                        model=explore.model_name,
+                                        explore=explore.name,
+                                        dimension=None,
+                                        sql="",
+                                        message=(
+                                            "Couldn't finish testing "
+                                            f"{explore.model_name}.{explore.name} "
+                                            "because queries repeatedly expired "
+                                            "in Looker."
+                                        ),
+                                        explore_url=query.explore_url,
+                                    )
+                                )
+                            query_slot.release()
+                            queries_to_run.task_done()
+                        else:
+                            await running_queries.put(task_id)
+
                     else:
                         # Query still running, put the task back on the queue
                         await running_queries.put(task_id)
@@ -405,7 +519,7 @@ class SqlValidator:
             while not queries_to_run.empty():
                 logger.debug("Waiting for the queries_to_run queue to clear")
                 # _run_query can get bogged down waiting for query slots, so free them
-                if query_slot.locked:
+                if query_slot.locked():
                     query_slot.release()
                 await asyncio.sleep(1)
             raise
